@@ -1,9 +1,3 @@
-"""
-Distortion functions to implement spectral risk measures
-
-May 2019: added capped log linear and tt distortions
-"""
-
 import numpy as np
 import matplotlib.pyplot as plt
 import scipy.stats as ss
@@ -11,43 +5,43 @@ from scipy.interpolate import interp1d
 from scipy.spatial import ConvexHull
 from io import StringIO
 import pandas as pd
-from . utils import axiter_factory, suptitle_and_tight
 from textwrap import fill
 import logging
 
-logger = logging.getLogger('aggregate')
+logger = logging.getLogger(__name__)
+
 
 class Distortion(object):
     """
-    handles everything to do with distortion functions
+    Creation and management of distortion functions.
 
+    0.9.4: renamed roe to ccoc, but kept creator with roe for backwards compatibility.
+    Oct 2022: renamed wtdtvar to bitvar, but kept ...
 
     """
+
     # make these (mostly) immutable...avoid changing by mistake
-    _available_distortions_ = ('ph', 'wang', 'cll', 'lep', 'ly', 'clin', 'dual', 'roe', 'tvar', 'wtdtvar', 'convex', 'tt')
+    _available_distortions_ = ('ph', 'wang', 'cll', 'lep', 'ly', 'clin', 'dual', 'ccoc', 'tvar',
+                               'bitvar', 'convex', 'tt')
     _has_mass_ = ('ly', 'clin', 'lep', 'roe')
-    _med_names_ = ("Prop Hzrd", "Wang", 'Capd Loglin', "Lev Equiv",
-                    "Lin Yield", "Capped Linear", "Dual Mom", "Const ROE", "Tail VaR", 'Wtd TVaR', "Convex Env", 'Wang-tt')
+    _med_names_ = ("Prop Hzrd", "Wang", 'Capd Loglin', "Lev Equiv", "Lin Yield", "Capped Linear", "Dual Mom",
+                   'Const CoC', "Tail VaR", 'BiTVaR', "Convex Env", 'Wang-tt')
     _long_names_ = ("Proportional Hazard", "Wang-normal", 'Capped Loglinear', "Leverage Equivalent Pricing",
-                    "Linear Yield", "Capped Linear", "Dual Moment", "Constant ROE", "Tail VaR", 'Weighted TVaR',
-                    "Convex Envelope", 'Wang-tt')
+                    "Linear Yield", "Capped Linear", "Dual Moment", "Constant CoC", "Tail VaR",
+                    'BiTVaR', "Convex Envelope", 'Wang-tt')
     # TODO fix examples!
-    # _available_distortions_ = ('ph', 'wang', 'cll', 'lep',  'ly', 'clin', 'dual', 'roe', 'tvar', 'wtdtvar,  'convex')
-    _eg_param_1_ =              (.9,     .1,      .9,    0.25,  0.8,   1.1,   1.5,    .1,    0.15,     .15)
-    _eg_param_2_ =              (.5,     .75,     .5,    0.5,   1.5,   1.8,     3,    .25,    0.5,      .5)
+    # _available_distortions_ = ('ph', 'wang', 'cll', 'lep',  'ly', 'clin', 'dual', 'ccoc', 'tvar', 'wtdtvar,  'convex')
+    _eg_param_1_ =              (.9,     .1,      .9,    0.25,  0.8,   1.1,   1.5,   .1,     0.15,     .15)
+    _eg_param_2_ =              (.5,     .75,     .5,    0.5,   1.5,   1.8,     3,   .25,    0.50,      .5)
     # _distortion_names_ = dict(zip(_available_distortions_, _med_names_))
     _distortion_names_ = dict(zip(_available_distortions_, _long_names_))
     renamer = _distortion_names_
 
-    # @property
-    # @staticmethod
-    # def renamer():
-    #     return _distortion_names_
 
     @classmethod
     def available_distortions(cls, pricing=True, strict=True):
         """
-        list of the available distortions
+        List of the available distortions.
 
         :param pricing: only return list suitable for pricing, excludes tvar and convex
         :param strict: only include those without mass at zero  (pricing only)
@@ -63,9 +57,9 @@ class Distortion(object):
 
     def __init__(self, name, shape, r0=0.0, df=None, col_x='', col_y='', display_name=''):
         """
-        create new distortion
+        Create a new distortion.
 
-        Tester:
+        Tester: ::
 
             ps = np.linspace(0, 1, 201)
             for dn in agg.Distortion.available_distortions(True):
@@ -84,7 +78,6 @@ class Distortion(object):
                 print("errors")
                 display(df.query(' abs(gg_inv - g_invg) > 1e-5'))
 
-
         :param name: name of an available distortion, call ``Distortion.available_distortions()`` for a list
         :param shape: float or [float, float]
         :param shape: shape parameter
@@ -94,6 +87,7 @@ class Distortion(object):
         :param col_y:
         :param display_name: over-ride name, useful for parameterized convex fix distributions
         """
+
         self._name = name
         self.shape = shape
         self.r0 = r0
@@ -193,9 +187,10 @@ class Distortion(object):
             def g_inv(x):
                 return np.where(x <= self.r0, 0, (x - self.r0) / sl)
 
-        elif self._name == 'roe':
+        elif self._name == 'roe' or self._name == 'ccoc':
             # constant roe = capped linear with shape = 1/(1+r), r0=r/(1+r)
             # r = target roe
+            self._name = 'ccoc'
             r = self.shape
             v = 1 / (1 + r)
             d = 1 - v
@@ -244,7 +239,7 @@ class Distortion(object):
             def g_inv(y):
                 return 1 - (1 - y)**q
 
-        elif self._name == 'wtdtvar':
+        elif self._name == 'wtdtvar' or self._name == 'bitvar':
             # weighted tvar, df = p0 <p1, shape = weight on p1
             try:
                 p0, p1 = df
@@ -296,16 +291,13 @@ class Distortion(object):
 
         self.g = g
         self.g_inv = g_inv
+
         def g_dual(x):
             return 1 - self.g(x)
+
         self.g_dual = g_dual
 
     def __str__(self):
-        """
-        printable version of distortion
-
-        :return:
-        """
         if self.display_name != '':
             s = self.display_name
             return s
@@ -319,7 +311,7 @@ class Distortion(object):
             s += f', ({self.df[0]:.3f}/{self.df[1]:.3f})'
         elif self.has_mass:
             # don't show the mass for wtdtvar
-            s += f', {self.r0:.3f}'
+            s += f', mass {self.mass:.3f}'
 
         return s
 
@@ -343,12 +335,13 @@ class Distortion(object):
 
     def plot(self, xs=None, n=101, both=True, ax=None, plot_points=True, scale='linear', c=None, **kwargs):
         """
-        quick plot of the distortion
+        Quick plot of the distortion
 
-        :param ax:
         :param xs:
         :param n:  length of vector is no xs
         :param both: True: plot g and ginv and add decorations, if False just g and no trimmings
+        :param ax:
+        :param plot_points:
         :param scale: linear as usual or return plots -log(gs)  vs -logs and inverts both scales
         :param kwargs:  passed to plot
         :return:
@@ -356,6 +349,10 @@ class Distortion(object):
 
         assert scale in ['linear', 'return']
 
+        if scale == 'return' and n == 101:
+            # default not enough for return
+            n = 10001
+
         if xs is None:
             xs = np.linspace(0, 1, n)
 
@@ -375,7 +372,7 @@ class Distortion(object):
             ax.plot(xs, y1, c='C0', label='$g$', **kwargs)
             if both:
                 ax.plot(xs, y2, c='C1', label='$g^{-1}$', **kwargs)
-            ax.set(xscale='log', yscale='log', xlim=[1/2000, 1], ylim=[1/2000, 1])
+            ax.set(xscale='log', yscale='log', xlim=[1/5000, 1], ylim=[1/5000, 1])
             ax.plot(xs, xs, lw=0.5, color='black', alpha=0.5)
 
         if self._name == 'convex' and plot_points:
@@ -392,16 +389,14 @@ class Distortion(object):
             elif scale == 'return':
                 ax.scatter(x=1/self.df[self.col_x], y=1/self.df[self.col_y], marker='.', s=15, color=c, alpha=alpha)
 
-        ax.grid(linewidth='0.25')
-
-        ax.set(title=fill(str(self), 20), aspect='equal')
+        ax.set(title=fill(str(self), 20), aspect='equal');
 
         return ax
 
     @classmethod
     def test(cls, r0=0.035, df=[0.0, .9]):
         """
-        tester: make some nice plots
+        Tester: make some nice plots of available distortions.
 
         :return:
         """
@@ -434,53 +429,43 @@ class Distortion(object):
 
         # tidy up
         for ax in axiter0:
-            f.delaxes(ax)
+            f0.delaxes(ax)
         for ax in axiter1:
-            f.delaxes(ax)
+            f1.delaxes(ax)
 
         f0.suptitle('Example Distortion Functions - Linear Scale')
         f1.suptitle('Example Distortion Functions - Return Scale')
 
     @staticmethod
-    def distortions_from_params(params, index, r0=0.025, df=5.5, plot=True, axiter=None, pricing=True, strict=True):
+    def distortions_from_params(params, index, r0=0.025, df=5.5, pricing=True, strict=True):
         """
-        make set of dist funs and inverses from params, output of port.calibrate_distortions
-        params must just have one row for each method and be in the output format of cal_dist
+        Make set of dist funs and inverses from params, output of port.calibrate_distortions.
+        params must just have one row for each method and be in the output format of cal_dist.
+
+        Called by Portfolio.
 
-        :param plot:
         :param index:
-        :param r0: min rol parameters
         :param params: dataframe such that params[index, :] has a [lep, param] etc.
-        pricing=True, strict=True: which distortions to allow
-        df for t distribution
+               pricing=True, strict=True: which distortions to allow
+               df for t distribution
+        :param r0: min rol parameters
+        :param strict:
+        :param pricing:
         :return:
         """
+
         temp = params.loc[index, :]
         dists = {}
         for dn in Distortion.available_distortions(pricing=pricing, strict=strict):
             param = float(temp.loc[dn, 'param'])
             dists[dn] = Distortion(name=dn, shape=param, r0=r0, df=df)
 
-        if plot:
-            axiter = axiter_factory(axiter, len(dists))
-            # f, axs = plt.subplots(2, 3, figsize=(8, 6))
-            # it = iter(axs.flatten())
-            for dn in Distortion.available_distortions(pricing=pricing, strict=strict):
-                dists[dn].plot(ax=next(axiter))
-            try:
-                axiter.tidy()
-                plt.tight_layout()
-            except:
-                # fails if axiter is just an iteration of axis elements
-                # assume then that constrained_layout =True so no tight layout
-                pass
-
-        return dists  # [g_lep, g_ph, g_wang, g_ly, g_clin]
+        return dists
 
     @staticmethod
     def convex_example(source='bond'):
         """
-        example convex distortion using data from https://www.bis.org/publ/qtrpdf/r_qt0312e.pdf
+        Example convex distortion using data from https://www.bis.org/publ/qtrpdf/r_qt0312e.pdf.
 
         :param source: bond gives a bond yield curve example, cat gives cat bond / cat reinsurance pricing based example
         :return:
@@ -524,12 +509,12 @@ class Distortion(object):
     @staticmethod
     def bagged_distortion(data, proportion, samples, display_name="", random_state=None):
         """
-        make a distortion by bootstrap aggregation (Bagging) resampling, taking the convex envelope,
-        and averaging from data
+        Make a distortion by bootstrap aggregation (Bagging) resampling, taking the convex envelope,
+        and averaging from data.
 
-        Each sample uses proportion of the data
+        Each sample uses proportion of the data.
 
-        data has columns just two columns: EL and Spread
+        Data must have  two columns: EL and Spread
 
         :param data:
         :param proportion: proportion of data for each sample
@@ -560,8 +545,8 @@ class Distortion(object):
     @staticmethod
     def average_distortion(data, display_name, n=201, el_col='EL', spread_col='Spread'):
         """
-        create average distortion from (s, g(s)) pairs. Each point defines a wtdTVaR with
-        p=s and p=1 points
+        Create average distortion from (s, g(s)) pairs. Each point defines a wtdTVaR with
+        p=s and p=1 points.
 
         :param data:
         :param display_name:
@@ -589,7 +574,7 @@ class Distortion(object):
     @staticmethod
     def wtd_tvar(ps, wts, display_name='', details=False):
         """
-        a careful version of wtd tvar with knots at ps and wts
+        A careful version of wtd tvar with knots at ps and wts.
 
         :param ps:
         :param wts:
@@ -619,7 +604,7 @@ class Distortion(object):
     @staticmethod
     def s_gs_distortion(s, gs, display_name=''):
         """
-        make a convex envelope distortion from {s, g(s)} points
+        Make a convex envelope distortion from {s, g(s)} points.
 
         :param s: iterable (can be converted into numpy.array
         :param gs:
